from math import pi
from ngsolve import *
from netgen.csg import OrthoBrick, Pnt

LevelsetExamples = {
    # geometry from 'Dziuk, Elliott, Finite element methods for surface PDEs, Acta Numerica, 2013', pp. 373-374:
    "cheese" : (sqrt((x*x-1)*(x*x-1)+(y*y-1)*(y*y-1)+(z*z-1)*(z*z-1)+(x*x+y*y-4)*(x*x+y*y-4)+(x*x+z*z-4)*(x*x+z*z-4)+(y*y+z*z-4)*(y*y+z*z-4))-4).Compile(),
    # geometry from 'Dziuk, Elliott, Finite element methods for surface PDEs, Acta Numerica, 2013', pp. 318-319:
    "dziukelliott" : sqrt(0.25*x*x+y*y+4.0*z*z/((1+0.5*sin(pi*x))*(1+0.5*sin(pi*x))))-1.0,
    # geometry from 'Dziuk, Finite elements for the beltrami operator on arbitrary surfaces':
    "dziuk88" : sqrt( (x-z*z)*(x-z*z)+y*y+z*z) - 1.0,
    # the unit sphere
    "sphere" : sqrt(x*x+y*y+z*z)-1.0,
    # torus with parameters as in 'Grande, Reusken, A higher order finite element method for partial differential euqations on surface, SINUM, 2016'
    "torus" : sqrt(z*z + (sqrt(x*x+y*y) - 1.0)*(sqrt(x*x+y*y) - 1.0)) - 0.6,
    # gyroid from Lehrenfeld, CMAME, 2016
    "gyroid" : cos(pi*x)*sin(pi*y)+cos(pi*y)*sin(pi*z)+cos(pi*z)*sin(pi*x),
}

BoundingBoxes = {
    "cheese" : OrthoBrick(Pnt(-2.5,-2.5,-2.5), Pnt(2.5,2.5,2.5)),
    "dziukelliott" : OrthoBrick(Pnt(-2.5,-1.5,-1.5), Pnt(2.5,1.5,1.5)),
    "dziuk88" : OrthoBrick(Pnt(-2,-2,-2), Pnt(2,2,2)),
    "sphere" : OrthoBrick(Pnt(-1.5,-1.5,-1.5), Pnt(1.5,1.5,1.5)),
    "torus" : OrthoBrick(Pnt(-2,-2,-2), Pnt(2,2,2)),
    "gyroid" : OrthoBrick(Pnt(-1,-1,-1), Pnt(1,1,1)),
}



from ngsolve import *
from time import sleep
from netgen.geom2d import unit_square
from netgen.geom2d import SplineGeometry
from netgen.meshing import MeshingParameters

from ngsolve.internal import *
from xfem import *
from math import pi


class LevelSetMeshAdaptation_Spacetime:
    """
Class to compute a proper mesh deformation to improve a piecewise (multi-) 
linear (in space!) level set approximation to obtain a higher order 
accurate approximation.

Computes a continuous function that describes a shift between points that 
are on the (P1-in-space ) approximated level set function and its higher 
order accurate approximation. The transformation is only applied on elements 
where a level value inside a certain interval (lower,upper) exists.

The result is a space-time finite element deformation (`deform`). For each 
fixed time the behavior is as for an `LevelSetMeshAdaptation` object, i.e.


1)phi_lin( Psi(x) ) = phi_h(x)

  with Psi(x) = x + d(x) qn(x) =: D(x)

for all x on 'cut' elements

with

  phi_h : self.lset_ho
    the higher order space-time level set function

  phi_lin : self.lset_p1
    the P1-in-space space-time level set function

  Psi : Id + self.deform
    the resulting deformation

  qn : self.qn
    normal direction field

This class holds its own members for the higher order and lower order
 (P1-in-space) space-time approximation of the level set function and 
 only depends on the input of a mesh and a CoefficientFunction 
 (the levelset function) (and options).

 A LevelSetMeshAdaptation_Spacetime can also be used as a context
 manager. In this case, the mesh deformation is applied on the mesh
 inside the context.
    """

    order_deform = 2
    order_qn = 2
    order_lset = 2


    @TimeFunction
    def __init__(self, mesh, order_space = 2, order_time = 1, lset_lower_bound = 0,
                 lset_upper_bound = 0, threshold = -1, discontinuous_qn = False, heapsize=1000000,periodic=False):

        """
The computed deformation depends on different options:

  order_space : int
    order of deformation GridFunction (ideally) order + 1 is the accuracy of the geometry
    approximation after applying the deformation on the mesh

  order_time : int
    order in time for space-time tensor-product spaces

  lset_lower_bound: float
    smallest relevant level set value to define the 'cut' elements where the mapping should be
    applied
    
  lset_upper_bound: float
    largest relevant level set value to define the 'cut' elements where the mapping should be
    applied
    
  threshold: float
    maximum (pointwise) value for d(x)/h in the mapping
      Psi(x) = x + d(x) qn(x)
    of the mesh transformation. A small value might be necessary if the geometry is only coarsely
    approximated to avoid irregular meshes after a corresponding mesh deformation.

  discontinuous_qn: boolean
    As an approximation for the normal direction we use n_h = nabla phi_h (gradient of higher order
    level set approximation) depending on the discontinuous_qn flag this normal field will be
    projected onto a continuous finite element space or not.

  eps_perturbation: float
    epsilon perturbation that is used to interpolate to P1

  heapsize : int
    heapsize for local computations.

  levelset : CoefficientFunction or None(default)
    If a level set function is prescribed the deformation is computed right away. Otherwise the 
    computation is triggered only at calls for `CalcDeformation`.
        """

        self.gf_to_project = []
        self.gf_to_project_tmp = [] # list for temporary copies
        
        self.order_deform = order_space
        self.order_qn = order_space
        self.order_lset = order_space
        self.order_time = order_time

        self.lset_lower_bound = lset_lower_bound
        self.lset_upper_bound = lset_upper_bound
        self.threshold = threshold
        self.periodic = periodic
        
        self.mesh = mesh
        self.v_ho = H1(mesh, order=self.order_lset)
        self.lset_ho_node = GridFunction (self.v_ho, "lset_ho_node")
        self.ndof_node = len(self.lset_ho_node.vec)

        if (discontinuous_qn):
            self.v_qn = L2(mesh, order=self.order_qn, dim=mesh.dim)
        else:
            self.v_qn = H1(mesh, order=self.order_qn, dim=mesh.dim)
        self.qn = GridFunction(self.v_qn, "qn")
    
        self.v_p1 = H1(mesh, order=1)
        self.lset_p1_node = GridFunction (self.v_p1, "lset_p1_node")
        self.ndof_node_p1 = len(self.lset_p1_node.vec)

        if self.periodic:
            self.v_def = Periodic(H1(mesh, order=self.order_deform, dim=mesh.dim))
        else:
            self.v_def = H1(mesh, order=self.order_deform, dim=mesh.dim)
        
        self.deform_node = GridFunction(self.v_def, "deform_node")
        self.heapsize = heapsize
        
        # Spacetime
        self.tfe = ScalarTimeFE(self.order_time) 
        
        self.v_ho_st = SpaceTimeFESpace(self.v_ho,self.tfe)
        self.lset_ho = GridFunction(self.v_ho_st)

        self.v_p1_st = SpaceTimeFESpace(self.v_p1,self.tfe)
        self.lset_p1 = GridFunction(self.v_p1_st)
        self.lset_p1_top = CreateTimeRestrictedGF(self.lset_p1,1.0)
        self.lset_p1_bottom = CreateTimeRestrictedGF(self.lset_p1,0.0)        
        
        self.v_def_st = SpaceTimeFESpace(self.v_def,self.tfe)
        self.deform = GridFunction(self.v_def_st)
        self.deform_top = CreateTimeRestrictedGF(self.deform,1.0)
        self.deform_bottom = CreateTimeRestrictedGF(self.deform,0.0)
        
        self.ci = CutInfo(mesh)
        
        self.hasneg_spacetime = BitArray(self.ci.GetElementsOfType(NEG))
        self.hasneg_spacetime[:] = False
        self.haspos_spacetime = BitArray(self.ci.GetElementsOfType(POS))
        self.haspos_spacetime[:] = False
        self.hasif_spacetime = BitArray(self.ci.GetElementsOfType(IF))
        self.hasif_spacetime[:] = False
        
        self.v_kappa_node = L2(mesh,order=0)
        self.v_kappa = SpaceTimeFESpace(self.v_kappa_node,self.tfe)
        self.kappa = GridFunction(self.v_kappa, "kappa")        

        self.deform_last_top = CreateTimeRestrictedGF(self.deform,1.0)

        class MeshDeformationContext:
            def __init__(self,lsetadap_st,type="std"):
                self.lsetadap_st = lsetadap_st
                self.type = type
            def __enter__(self):
                if self.type == "std":
                    self.lsetadap_st.mesh.SetDeformation(self.lsetadap_st.deform)
                    return self.lsetadap_st.lset_p1
                elif self.type == "top":   
                    self.lsetadap_st.mesh.SetDeformation(self.lsetadap_st.deform_top)
                    return self.lsetadap_st.lset_p1_top
                elif self.type == "bottom":   
                    self.lsetadap_st.mesh.SetDeformation(self.lsetadap_st.deform_bottom)
                    return self.lsetadap_st.lset_p1_bottom
                else:
                    raise Exception("no suitable deform type given")
            def __exit__(self, type, value, tb):
                self.lsetadap_st.mesh.UnsetDeformation()
        self.top = MeshDeformationContext(self,"top")
        self.bottom = MeshDeformationContext(self,"bottom")

        self.levelsetp1 = {INTERVAL : self.lset_p1, BOTTOM : self.lset_p1_bottom, TOP : self.lset_p1_top}
        self.deformation = {INTERVAL : self.deform, BOTTOM : self.deform_bottom, TOP : self.deform_top}

        
    def ProjectOnUpdate(self,gf,update_domain=None):
        """
When the LevelsetMeshAdaptation class generates a new deformation (due to 
a new level set function) all GridFunction that have been stored through
`ProjectOnUpdate` will be projected from the previous to the new mesh
by an essentially local projection (Oswald projection of the shifted 
evaluation)

  gfu : GridFunction (or list of GridFunctions)
    GridFunction(s) to store for later deformation updates.
        """      
        if isinstance(gf,list):
            self.gf_to_project.extend([(gfa, update_domain) for gfa in gf])
            self.gf_to_project_tmp.extend([GridFunction(gfa.space) for gfa in gf])
        else:
            self.gf_to_project.append((gf, update_domain))
            self.gf_to_project_tmp.append(GridFunction(gf.space))

    @TimeFunction
    def ProjectGFs(self):
        """
ProjectGFs projects all stored GridFunctions to the currect deformation.
This function is typically only called in the `CalcDeformation` unless
`CalcDeformation` is called with the argument `dont_project_gfs = True`.
        """      

        for (gf, update_domain),gftmp in zip(self.gf_to_project, self.gf_to_project_tmp):
            gftmp.vec.data = gf.vec
            if update_domain is None:
                if (gf.space.mesh.ne != self.deform_bottom.space.mesh.ne):
                    raise Exception("different meshes...")
                gf.Set(shifted_eval(gftmp, back = self.deform_last_top, forth = self.deform_bottom))
            else:
                gf.Set(shifted_eval(gftmp, back = self.deform_last_top, forth = self.deform_bottom), definedonelements=update_domain)
            #print("updated ", gf.name)

    @TimeFunction
    def interpol_ho(self,levelset):
        """
Internal function that is called inside `CalcDeformation`. It projects 
the space-time coefficient function to a space-time tensor product finite 
element function.
        """
        times = [xi for xi in self.v_ho_st.TimeFE_nodes()]
        for i,ti in enumerate(times):
            self.lset_ho_node.Set(fix_tref(levelset,ti))
            self.lset_ho.vec[i*self.ndof_node : (i+1)*self.ndof_node].data = self.lset_ho_node.vec[:]

    @TimeFunction
    def interpol_p1(self):
        """
Internal function that is called inside `CalcDeformation`. It projects 
the space-time tensor product finite element approximation lset_ho 
to a piecewise linear-in-space approximation through interpolation
(in space).
        """
        for i in range(self.order_time + 1):
            self.lset_ho_node.vec[:].data = self.lset_ho.vec[i*self.ndof_node : (i+1)*self.ndof_node]
            InterpolateToP1(self.lset_ho_node,self.lset_p1_node)
            self.lset_p1.vec[i*self.ndof_node_p1 : (i+1)*self.ndof_node_p1].data = self.lset_p1_node.vec[:]

    @TimeFunction
    def CalcDeformation(self, levelset, calc_kappa = False, dont_project_gfs = False):
        """
Compute the space-time mesh deformation, s.t. isolines on cut elements of lset_p1 (the piecewise 
linear-in-space tensor product approximation) are mapped towards the corresponding isolines of 
a given space-time function

Parameters:

levelset : CoefficientFunction
  The coefficient function that prescribes where the piecewise linear-in-space iso lines should 
  be mapped to.

calc_kappa : bool
  Compute the cut ratio of a spatial element as a space-time function. Populates `self.kappa`.

dont_project_gfs : bool
  Usually, all stored GridFunctions are projected on the new deformed mesh. This flag
  can be used to avoid the call of the projection step (`ProjectGFs`).
        """

        self.v_ho.Update()
        self.lset_ho_node.Update()
        self.v_p1.Update()
        self.lset_p1_node.Update()
        self.v_qn.Update()
        self.qn.Update()
        self.v_def.Update()
        self.deform_node.Update()
        self.v_kappa_node.Update()
        self.v_kappa.Update()
        self.kappa.Update()
        self.deform_last_top.Update()               
        self.interpol_ho(levelset)
        self.interpol_p1()
        RestrictGFInTime(spacetime_gf=self.deform,reference_time=1.0,space_gf=self.deform_last_top)   


        for i in  range(len(self.v_ho_st.TimeFE_nodes())):
            self.lset_p1_node.vec[:].data = self.lset_p1.vec[i*self.ndof_node_p1 : (i+1)*self.ndof_node_p1]
            if calc_kappa:
                self.ci.Update(self.lset_p1_node)
                self.kappa.vec[i*self.v_kappa_node.ndof : (i+1)*self.v_kappa_node.ndof].data = self.ci.GetCutRatios(VOL)
        
        
        self.ci.Update(self.lset_p1,time_order=self.order_time)
        self.hasneg_spacetime[:] = False
        self.hasneg_spacetime |= self.ci.GetElementsOfType(NEG)
        self.haspos_spacetime[:] = False
        self.haspos_spacetime |= self.ci.GetElementsOfType(POS)
        self.hasif_spacetime[:] = False
        self.hasif_spacetime |= self.ci.GetElementsOfType(IF)
        
        for i in range(self.order_time + 1):
            self.lset_ho_node.vec[:].data = self.lset_ho.vec[i*self.ndof_node : (i+1)*self.ndof_node]
            self.qn.Set(self.lset_ho_node.Deriv())
            self.lset_p1_node.vec[:].data = self.lset_p1.vec[i*self.ndof_node_p1 : (i+1)*self.ndof_node_p1]
            ProjectShift(self.lset_ho_node, self.lset_p1_node, self.deform_node, self.qn, self.hasif_spacetime, None, self.lset_lower_bound, 
                         self.lset_upper_bound, self.threshold, heapsize=self.heapsize)
            self.deform.vec[i*self.ndof_node : (i+1)*self.ndof_node].data = self.deform_node.vec[:]


        RestrictGFInTime(spacetime_gf=self.lset_p1,reference_time=0.0,space_gf=self.lset_p1_bottom)
        RestrictGFInTime(spacetime_gf=self.lset_p1,reference_time=1.0,space_gf=self.lset_p1_top)
        RestrictGFInTime(spacetime_gf=self.deform,reference_time=0.0,space_gf=self.deform_bottom)   
        RestrictGFInTime(spacetime_gf=self.deform,reference_time=1.0,space_gf=self.deform_top)   

        if not dont_project_gfs:
            self.ProjectGFs()
        return self.deform
            
    def __enter__(self):
      self.mesh.SetDeformation(self.deform)
      return self.lset_p1

    def __exit__(self, type, value, tb):
      self.mesh.UnsetDeformation()

    def levelset_domain(self, domain_type = IF, time_type = INTERVAL):
        """
Return a levelset_domain dictionary.
      Args:
          domain_type (DOMAIN_TYPE: NEG/POS/IF, optional): domain type for level set doamin. Defaults to IF.
          time_type (BOTTOM/TOP/INTERVAL): type of integral on the space-time domain

      Returns:
          dict: levelset domain
        """        
        return { "levelset" : self.levelsetp1[time_type], "domain_type" : domain_type}



    def Integrator(self, SymbolicFI, domain_type, time_type, form, time_order = None, definedonelements = None):
        """
Convenience function to construct space-time Symbolic(Cut)LFI/Symbolic(Cut)BFI

        Args:
            SymbolicFI (SymbolicLFI/SymbolicBFI): type of integrator (LFI vs. BFI)
            domain_type (DOMAIN_TYPE: NEG/POS/IF): domain type to integrate on
            time_type (BOTTOM/TOP/INTERVAL): type of integral on the space-time domain
            form (CoefficientFunction): integrand
            time_order (int, optional): integration order in time. Defaults to None.
            defineonelement (BitArray, optional): elements to define the linear form on. Defaults to None.
        """    
        if time_order == None:
            time_order = 2 * self.order_time
        
        fi = SymbolicFI(levelset_domain = self.levelset_domain(domain_type,time_type), 
                         form = form, 
                         time_order=time_order, 
                         deformation=self.deformation[time_type])
        if definedonelements != None:
            fi.SetDefinedOnElements(definedonelements)       
        return fi

    def LFI(self, domain_type, time_type, form, time_order = None, definedonelements = None):
        """
Convenience function to construct Symbolic(Cut)LFI based on the domain_type 
and the known space-time level set function.

        Args:
            domain_type (DOMAIN_TYPE: NEG/POS/IF): domain type to integrate on
            time_type (BOTTOM/TOP/INTERVAL): type of integral on the space-time domain
            form (CoefficientFunction): integrand
            time_order (int, optional): integration order in time. Defaults to None.
            defineonelement (BitArray, optional): elements to define the linear form on. Defaults to None.
        """    
        return self.Integrator(SymbolicLFI, domain_type, time_type, form, time_order, definedonelements)

    def BFI(self, domain_type, time_type, form, time_order = None, definedonelements = None):
        """
Convenience function to construct Symbolic(Cut)LFI based on the domain_type 
and the known space-time level set function.

        Args:
            domain_type (DOMAIN_TYPE: NEG/POS/IF): domain type to integrate on
            time_type (BOTTOM/TOP/INTERVAL): type of integral on the space-time domain
            form (CoefficientFunction): integrand
            time_order (int, optional): integration order in time. Defaults to None.
            defineonelement (BitArray, optional): elements to define the bilinear form on. Defaults to None.
        """    
        return self.Integrator(SymbolicBFI, domain_type, time_type, form, time_order, definedonelements)

    @TimeFunction
    def Integrate(self, domain_type, time_type, cf, order = 5, time_order = None):
        """
Convenience function to Integrate on cut space-time domain.

        Args:
            domain_type (DOMAIN_TYPE: NEG/POS/IF): domain type to integrate on
            time_type (BOTTOM/TOP/INTERVAL): type of integral on the space-time domain
            cf (CoefficientFunction): integrand
            order (int, optional): integration order in space. Defaults to 5.
            time_order (int, optional): integration order in time. Defaults to None.

        Returns:
            float: return value of (cut) integral
        """
        if time_order == None:
            time_order = 2 * self.order_time
        self.mesh.SetDeformation(self.deformation[time_type])
        fi = Integrate(levelset_domain = self.levelset_domain(domain_type,time_type), 
                       mesh = self.mesh, cf = cf, order = order, 
                       time_order=time_order)
        self.mesh.UnsetDeformation()
        return fi


    @TimeFunction
    def CalcMaxDistance(self, levelset, order=None, time_order=None, heapsize=None):
        """
Compute approximated distance between of the isoparametrically obtained geometry
(should be called in deformed state)

        Args:
            levelset (CoefficientFunction): implicit geometry description, 
              assumed to be a signed distance function.
            order (int, optional): integration order in space. Defaults to 5.
            time_order (int, optional): integration order in time. Defaults to None.
            heapsize (int, optional): heap size for local memory. Defaults to None.
        """
        if order == None:
          order = 2 * self.order_qn
        if time_order == None:
          time_order = 2 * self.order_time
        if heapsize == None:
          heapsize = self.heapsize
        lset_dom = {"levelset": self.lset_p1, "domain_type" : IF, "order": order, "time_order" : time_order}
        minv, maxv = IntegrationPointExtrema(lset_dom, self.mesh, levelset, heapsize=heapsize)
        return max(abs(minv),abs(maxv))



from ngsolve.la import BaseMatrix
from ngsolve import Projector, Norm
from ngsolve.krylovspace import CG
from xfem import *
from ngsolve import *
import ngsolve
'''
Below we implement a generic MultiGrid class based on the following (yet to be defined) objects:
* prolongation : prolongation from last to current mesh level 
* smoothers : smoothers corresponding to each mesh level
* coarsegridsolver : BaseMatrix for the coarse(st) level representing an (approx.) inverse
* mats : matrices corresponding to each mesh level
* nu : number of smoothing steps

The MultiGrid class is a BaseMatrix. Hence, it essentially implements the Mult operation.
'''
class MultiGridCL(BaseMatrix):
	def __init__(self,**kwargs):
	    super(MultiGridCL,self).__init__()
	    self.nu     = kwargs.get('nu',2)            
	    self.l      = kwargs['level'] # maximum number of level
	    self.prol   = kwargs['prol']
	    self.mats   = kwargs['matrices']
	    self.smoothers  = kwargs['smoothers']
	    self.inv    = kwargs["coarsegridsolver"]
	    self.errors     = [self.mats[-1].CreateColVector() for i in range(self.l+1)]
	    self.defects    = [self.mats[-1].CreateColVector() for i in range(self.l+1)]
	    self.level = self.l # current level
	
	def Mult(self,rhs,u):            
	    error   = self.errors[self.level]
	    defect  = self.defects[self.level]
	
	    if self.level == 0 :
	        u.data = self.inv*rhs
	        return
	#smoothing
	    self.smoothers[self.level].Smooth(u,rhs,self.nu)
	#compute defect
	    defect.data = rhs - self.mats[self.level]*u
	#restrict defect
	    self.prol.Restrict(self.level,defect)
	    error[:] = 0
	    #v-cycle
	    self.level -= 1
	    self.Mult(defect,error)
	    self.level += 1
	    self.prol.Prolongate(self.level,error)
	    u.data += error
	#smoothing
	    self.smoothers[self.level].SmoothBack(u,rhs,self.nu)
	    return u
	
	def Height(self):
	    return self.mats[-1].height
	def Width(self):
	    return self.mats[-1].width



class CutFemSmoother:     
    def __init__(self,**kwargs):
        CutVh, ci, self.a  = kwargs["CutVh"],kwargs["ci"], kwargs["a"]
        self.ifsolver = kwargs["ifsolver"]
        ifdofs = GetDofsOfElements(CutVh, ci.GetElementsOfType(IF)) & CutVh.FreeDofs()
        #self.proj = Projector(mask=self.ifdofs,range=True)
        ifdofslst = [ [i] for i in range(len(ifdofs)) if ifdofs[i]==True ]
        if self.ifsolver == "direct":
            self.ifpre = self.a.mat.Inverse(ifdofs, inverse="sparsecholesky")
        elif self.ifsolver == "cg":
            print("cg if solver")
            self.ifpre = self.a.mat.CreateBlockSmoother(ifdofslst)
        else:
            pass
            
        if "blocks" in kwargs:
            self.preJ = self.a.mat.CreateBlockSmoother( kwargs["blocks"] )
        else:
            self.preJ = self.a.mat.CreateSmoother( CutVh.FreeDofs() )
        
        self.ifcorr_only_once = kwargs.get('ifcorr_only_once',True)

    def InterfaceCorrection(self, u, rhs,k,nu):
        if self.ifsolver == None:
            return
        if self.ifcorr_only_once and k < nu-1:
            return
        update = u.CreateVector()
        # this needs to be more efficient 
        # with help of ifdofs ...
        update.data = self.a.mat * u - rhs
        cgoutput = u.CreateVector()

        E = Embedding(len(rhs), IntRange(0,self.a.mat.height))
        
        cgoutput.data = CG(E @ self.a.mat @ E.T,
                           update,
                           pre=E @ self.ifpre @ E.T,
                           tol=1e-2,printrates=False)
        #u.data -= self.inv * update
        u.data -= cgoutput

    def Smooth( self, u, rhs, nu ):
        for k in range(nu):
            self.preJ.Smooth(u, rhs)        
            self.InterfaceCorrection(u,rhs,k,nu)
            
    def SmoothBack( self, u, rhs, nu ):
        for k in reversed(range(nu)):
            self.InterfaceCorrection(u,rhs,k,nu)
            self.preJ.SmoothBack(u, rhs)



class LinearMGIterator(BaseMatrix):
    def __init__(self, **kwargs):
        super(LinearMGIterator,self).__init__()
        #self.nref = kwargs['nref']
        self.ifsolver = kwargs['ifsolver']
        #AssembleProblem = kwargs['assemblefct'] #not used
        ProlType = kwargs.get('ProlType',P1Prolongation)
        self.nu = kwargs.get('nu',2)
        mesh = kwargs['mesh']
        
        self.tol=kwargs.get("tol",1e-6)
        self.maxit=kwargs.get("maxit",20)
        self.printinfo=kwargs.get("printinfo",True)
        
        ci, lsetp1 = kwargs["ci"], kwargs["lsetp1"]
        #create coarse level bilinear/linear form
        a = kwargs["a"]
        self.CutVh = a.space
        
        # levelset fe space
        #VhL = lsetp1.space

        ci = CutInfo(mesh, lsetp1)
        
        print( type(self.CutVh) )
        #check for compound space
        if isinstance(self.CutVh,ProductSpace):
            print('creating compound cut prolongation')
            self.CutProl = CompoundProlongation( self.CutVh )
            # construct prolongation for neg/pos space
            self.CutProl.AddProlongation( ProlType(mesh) )
            self.CutProl.AddProlongation( ProlType(mesh) )
        else:
            print('creating single phase cut prolongation')
            self.CutProl = ProlType(mesh)
        
        self.CutProl.Update(self.CutVh)

        #self.mats = {}
        self.mats = [a.mat]
        #list of smoothers
        self.smoothers = [None]
        #coarse grid solver
        self.coarseinv = a.mat.Inverse(self.CutVh.FreeDofs(), inverse="sparsecholesky")
        self.MGpre = self.coarseinv
    
    def Update(self, a, CutVh, ci):
        self.CutProl.Update(CutVh)
        self.mats.append(a.mat)
        self.CutVh = CutVh
        current_smoother = CutFemSmoother(a=a, CutVh=CutVh, ci=ci, ifsolver=self.ifsolver,
                                                 ifcorr_only_once = True)
        self.smoothers.append(current_smoother)
        # create/update multigrid solver
        self.MGpre = MultiGridCL ( level=len(self.mats)-1, prol=self.CutProl, 
                                   matrices=self.mats, smoothers=self.smoothers,
                                   nu=self.nu,
                                   coarsegridsolver=self.coarseinv)        
        
    def createVec(self):
        return self.mats[-1].CreateColVector()

    def getSpaceDim(self):
        print('-----------------------')
        print(type(self.CutVh))
        #if( type(self.CutVh) == ngsolve.comp.CompoundFESpace ):
        if( type(self.CutVh) == ngsolve.comp.CompoundFESpace or type(self.CutVh) == ngsolve.comp.FESpace):
            numspaces = len ( self.CutVh.components )
            return [ self.CutVh.components[i].ndof for i in range(numspaces) ]
        else:
            return [self.CutVh.ndof]
    
    
    def Mult(self,rhs,usol):            
        usol[:] = 0.

        res = usol.CreateVector()
        projres = usol.CreateVector()
        
        proj = Projector(mask=self.CutVh.FreeDofs(),range=True) 
        fnew = rhs #f.vec.data     

        projres.data = proj*fnew #f.vec
        normf = Norm(projres)

        for it in range(self.maxit):
            usol.data = self.MGpre*fnew
            res.data = fnew - self.mats[-1]*usol
            projres.data = proj*res
            res_norm = Norm(projres) / normf
            if self.printinfo:
                print("it =", it+1, " ||res||_2 =", res_norm)
            if res_norm < self.tol:
                break        

        #udraw = GridFunction( self.CutVh )
        #udraw.components[1].Set( solution[1], BND )
        #udraw.vec.data += usol.vec

        return usol

    def Height(self):
        return self.mats[-1].height
    def Width(self):
        return self.mats[-1].height

def VertPatches(fes,mesh):
    blocks = []
    freedofs = fes.FreeDofs()
    for v in mesh.vertices:
        vdofs = set()
        for el in mesh[v].elements:
            #print( fes.GetDofNrs(el) )
            #input('w')
            vdofs |= set(d for d in fes.GetDofNrs(el) if (freedofs[abs(d)] and d >=0) )
        #print(vdofs)
        blocks.append (vdofs)
    return blocks

def ElemPatches(fes):
    blocks = []
    freedofs = fes.FreeDofs()
    for el in fes.Elements():
        eldofs = set(d for d in fes.GetDofNrs(el) if (freedofs[abs(d)] and d >=0) )        
        blocks.append (eldofs)
    return blocks

def EdgePatches(fes,mesh):
    blocks = []
    freedofs = fes.FreeDofs()
    for edge in mesh.edges:
        edofs = fes.GetDofNrs(edge)

        if( len(edofs) == 0 ): continue
        
        # print('found edge dof: ', edofs)
        elcntr = 0
        edofs = set()
        for el in mesh[edge].elements:
            elcntr += 1
            # print('  with element ', elcntr, ' el: ', fes.GetDofNrs(el) )
            # print('  with element ', elcntr, ' el: ', el.nr )
            edofs |= set( d for d in fes.GetDofNrs(el) if(freedofs[abs(d)] and d>=0) )
    
        # input('next edge')
        blocks.append (edofs)
    return blocks


class P2TwoGridCL(BaseMatrix):
    def __init__(self,**kwargs):
        super(P2TwoGridCL,self).__init__()
        self.a = kwargs['a']
        #self.f = kwargs['f']
        self.fes = kwargs['fes']
        
        self.patchtype = kwargs.get('patchtype','edge')

        mesh = kwargs['mesh']

        if(self.patchtype == 'edge'):
            blocks = EdgePatches(self.fes,mesh)
        elif self.patchtype == 'vert':
            blocks = VertPatches(self.fes,mesh)
        elif self.patchtype == 'elem':
            blocks = ElemPatches(self.fes)
        else:
            raise ValueError("Unknown patchtype for block smoothing!\n choose between 'edge', 'vert' and 'elem'")
        
        self.ifsolver = kwargs['ifsolver']
        self.blockjac = CutFemSmoother( a=self.a, CutVh=self.fes, 
                                        ci=kwargs["ci"], blocks=blocks, ifsolver=self.ifsolver,
                                        ifcorr_only_once=False )

        # we need a linear multigrid as coarse grid correction
        self.linmgiter = kwargs['linmgiter']
        self.maxit = kwargs.get('maxit',20)
        self.tol = kwargs.get('tol',1e-6)
        self.nu = kwargs.get( 'nu',3 )

    def Mult(self,rhs,usol):
        proj = Projector(mask=self.fes.FreeDofs(),range=True)
        usol[:] = 0.
        #
        projres = rhs.CreateVector()
        projres.data = proj*rhs
        normf = Norm(projres)
        #
        LinDof = self.linmgiter.getSpaceDim()
        
        if( type(self.fes) == ngsolve.comp.CompoundFESpace or type(self.fes) == ngsolve.comp.FESpace):
            numspaces = len( self.fes.components )
            HOdof = [ self.fes.components[i].ndof for i in range(numspaces) ]
        else:
            HOdof = [ self.fes.ndof ]
        #
        print('lindof, hdof: ')
        print(LinDof,HOdof)
        oldres = Norm(projres)

        for it in range(self.maxit):
            # initial smoothing 
            self.blockjac.Smooth(usol, rhs, self.nu)

            res = usol.CreateVector()
            # perform multiplication more efficiently ...
            # only 'upper' part (vertex dofs) needed 
            res.data = rhs -self.a.mat * usol 
            #eliminate boundary condition
            projres.data = proj*res           

            mgrhs = self.linmgiter.createVec()            
            # copy data from two domains 
            # ... should be done in combination with vertexdofs ...
            #mgrhs.Range(0, LinDof[0]).data = projres.Range(0, LinDof[0] )            
            #mgrhs.Range( LinDof[0], len(mgrhs) ).data = projres.Range(HOdof[0], HOdof[0] + LinDof[1] )
            
            mgrhs.Range(0, LinDof[0]).data = projres.Range(0, LinDof[0] )
            if( len(LinDof) > 1 ):
                offsetc = 0
                offsetf = 0
                for i in range( len(LinDof) -1 ):
                    offsetc += LinDof[i]
                    offsetf += HOdof[i]
                    mgrhs.Range( offsetc, offsetc + LinDof[i+1] ).data = projres.Range(offsetf, offsetf + LinDof[i+1] )
            
            # coarse grid correction: multigrid on linears
            cup = mgrhs.CreateVector()
            cup.data = self.linmgiter*mgrhs #maxit=1,printinfo=False)

            # update solution
            # ... should be done in combination with vertexdofs ...
            
            usol.Range(0, LinDof[0] ).data += cup.Range(0, LinDof[0])
            if( len(LinDof) > 1 ):
                offsetc = 0
                offsetf = 0
                for i in range( len(LinDof) -1 ):
                    offsetc += LinDof[i]
                    offsetf += HOdof[i]
                    usol.Range( offsetf, offsetf + LinDof[i+1] ).data += cup.Range( offsetc, offsetc + LinDof[i+1] )
            '''
            usol.Range(0, LinDof[0] ).data += cup.Range(0, LinDof[0])
            usol.Range( HOdof[0], HOdof[0] + LinDof[1]).data += cup.Range( LinDof[0], len(mgrhs) )
            '''

            # compute residual for measurement
            res.data = rhs - self.a.mat*usol
            projres.data = proj*res
            res_norm = Norm(projres)
            
            print("tg-it =", it+1, "\t ||res||_2 = {0:.2E}".format(res_norm), "\t reduction: {0:.2f}".format(res_norm/oldres))
            if res_norm < self.tol*normf:
                break
            oldres = res_norm
        
        return usol

    
    # def iterate(self,nu=2,maxit=20,tol=1e-6):
        
    #     proj = Projector(mask=self.fes.FreeDofs(),range=True)
    #     usol = GridFunction(self.fes)
    #     usol.vec[:] = 0.

    #     projres = usol.vec.CreateVector()
    #     projres.data = proj*self.f.vec
    #     normf = Norm(projres)

    #     LinDof = self.mgiter.getSpaceDim()
    #     HOdof  = [ self.fes.components[i].ndof for i in [0,1] ]

    #     oldres = Norm(projres)

    #     for it in range(maxit):
    #         # initial smoothing 
    #         #for sm in range(nu):
    #         self.blockjac.Smooth(usol.vec, self.f.vec,nu)

    #         res = usol.vec.CreateVector()
    #         # perform multiplication more efficiently ...
    #         # only 'upper' part (vertex dofs) needed 
    #         res.data = self.f.vec -self.a.mat * usol.vec
            

    #         mgrhs = self.mgiter.createVec()
    #         # copy data from two domains 
    #         # ... should be done in combination with vertexdofs ...
    #         mgrhs.Range(0, LinDof[0]).data = projres.Range(0, LinDof[0] )
    #         mgrhs.Range( LinDof[0], len(mgrhs) ).data = projres.Range(HOdof[0], HOdof[0] + LinDof[1] )
    #         #print('lens: ', len(mgrhs), LinDof[0], LinDof[1], LinDof[0]+LinDof[1])

    #         # coarse grid correction: multigrid on linears
    #         cup = self.mgiter.iterate(mgrhs,maxit=1,printinfo=False)

    #         # update solution
    #         # ... should be done in combination with vertexdofs ...
    #         usol.vec.Range(0, LinDof[0] ).data += cup.vec.Range(0, LinDof[0])
    #         usol.vec.Range( HOdof[0], HOdof[0] + LinDof[1]).data += cup.vec.Range( LinDof[0], len(mgrhs) )

    #         # compute residual for measurement
    #         res.data = self.f.vec - self.a.mat*usol.vec
    #         projres.data = proj*res
    #         res_norm = Norm(projres)
            
    #         print("tg-it =", it+1, "\t ||res||_2 = {0:.2E}".format(res_norm), "\t reduction: {0:.2f}".format(res_norm/oldres))
    #         if res_norm < tol*normf:
    #             break
    #         oldres = res_norm

    #     udraw = GridFunction( self.fes )
    #     udraw.components[1].Set( solution[1], BND )
    #     udraw.vec.data += usol.vec
    #     return udraw
